from typing import Optional

from qualibrate import ExecutionParameters
from qualibrate.models.run_summary.base import BaseRunSummary

__all__ = ["GraphRunSummary"]


class GraphRunSummary(BaseRunSummary):
    parameters: Optional[ExecutionParameters] = None



import traceback
from collections.abc import Mapping, Sequence
from datetime import datetime
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Optional,
    TypeVar,
    cast,
)

import networkx as nx
from pydantic import create_model

from qualibrate.models.node_status import NodeStatus
from qualibrate.models.outcome import Outcome
from qualibrate.models.run_mode import RunModes
from qualibrate.models.run_summary.base import BaseRunSummary
from qualibrate.models.run_summary.graph import GraphRunSummary
from qualibrate.models.run_summary.run_error import RunError
from qualibrate.parameters import (
    ExecutionParameters,
    GraphParameters,
    NodeParameters,
    NodesParameters,
)
from qualibrate.q_runnnable import (
    QRunnable,
    file_is_calibration_instance,
    run_modes_ctx,
)
from qualibrate.qualibration_node import QualibrationNode
from qualibrate.utils.exceptions import StopInspection
from qualibrate.utils.logger_m import logger
from qualibrate.utils.read_files import get_module_name, import_from_path
from qualibrate.utils.type_protocols import TargetType

if TYPE_CHECKING:
    from qualibrate.orchestration.qualibration_orchestrator import (
        QualibrationOrchestrator,
    )

__all__ = ["QGraphBaseType", "QualibrationGraph", "NodeTypeVar"]

NodeTypeVar = TypeVar("NodeTypeVar", bound=QualibrationNode[NodeParameters])
GraphCreateParametersType = GraphParameters
GraphRunParametersType = ExecutionParameters
QGraphBaseType = QRunnable[GraphCreateParametersType, GraphRunParametersType]


class QualibrationGraph(
    QRunnable[GraphCreateParametersType, GraphRunParametersType],
    Generic[NodeTypeVar],
):
    """
    Represents a graph of nodes for calibration purposes.

    The `QualibrationGraph` class manages a directed graph composed of nodes
    (`QualibrationNode`) and containing reference to orchestrator for executing
    the entire graph. It supports defining connectivity, creating parameters for
    execution, and serializing the graph's state.

    Args:
        name (str): Name of the graph.
        parameters (GraphCreateParametersType): Parameters for creating the
            graph.
        nodes (Mapping[str, QualibrationNode]): A mapping of node names to
            nodes.
        connectivity (Sequence[tuple[str, str]]): Adjacency list representing
            node connectivity as pairs of node names.
        orchestrator (Optional[QualibrationOrchestrator]): Orchestrator for the
            graph. Defaults to None.
        description (Optional[str]): Description of the graph. Defaults to None.
        modes (Optional[RunModes]): Modes to configure the graph behavior.
            Defaults to None.

    Raises:
        StopInspection: If the graph is instantiated in inspection mode.
    """

    STATUS_FIELD = "status"
    _node_init_args = {STATUS_FIELD: NodeStatus.pending, "retries": 0}

    def __init__(
        self,
        name: str,
        parameters: GraphCreateParametersType,
        nodes: Mapping[str, NodeTypeVar],
        connectivity: Sequence[tuple[str, str]],
        orchestrator: Optional["QualibrationOrchestrator[NodeTypeVar]"] = None,
        description: Optional[str] = None,
        *,
        modes: Optional[RunModes] = None,
    ):
        super().__init__(name, parameters, description=description, modes=modes)
        self._nodes = self._validate_nodes_names_mapping(nodes)
        self._connectivity = connectivity
        self._graph: nx.DiGraph[NodeTypeVar] = nx.DiGraph()
        self._orchestrator = orchestrator
        self._initial_targets: Sequence[TargetType] = []
        self._add_nodes_and_connections()
        self.full_parameters_class = self._build_parameters_class()
        self.full_parameters: GraphRunParametersType = (
            self.full_parameters_class()
        )

        if self.modes.inspection:
            raise StopInspection(
                "Graph instantiated in inspection mode", instance=self
            )

    def _add_nodes_and_connections(self) -> None:
        """
        Adds nodes and their connections to the internal graph representation.

        This method iterates over the registered nodes and connectivity data to:
        - Add nodes to the graph based on their names.
        - Add edges (connections) between nodes if both nodes exist.

        Raises:
            ValueError: If a connection references a node that has not been
                registered. The error message includes the offending node name
                and the available node names.
        """

        for node_name in self._nodes:
            self._add_node_by_name(node_name)
        for v_name, x_name in self._connectivity:
            try:
                v = self._get_qnode_or_error(v_name)
                x = self._get_qnode_or_error(x_name)
            except ValueError as ex:
                issued_node_name = (
                    ex.args[1] if len(ex.args) > 1 else f"{v_name} or {x_name}"
                )
                raise ValueError(
                    f'Error creating QualibrationGraph "{self.name}": Could '
                    f"not add connection ({v_name}, {x_name}) because node "
                    f'with name "{issued_node_name}" has not been registered. '
                    f"Available node names: {tuple(self._nodes.keys())}"
                ) from ex
            if not self._graph.has_edge(v, x):
                self._graph.add_edge(v, x)

    @staticmethod
    def _validate_nodes_names_mapping(
        nodes: Mapping[str, NodeTypeVar],
    ) -> Mapping[str, NodeTypeVar]:
        """
        Validates the mapping of nodes and ensures unique names.

        If the provided node names do not match the actual node names, a copy
        of the node is created with the correct name.

        Args:
            nodes (Mapping[str, QualibrationNode]): Mapping of node names to
                nodes.

        Returns:
            Mapping[str, QualibrationNode]: A valid mapping of node names to
                nodes.
        """
        new_nodes = {}
        for name, node in nodes.items():
            if name != node.name:
                node = cast(NodeTypeVar, node.copy(name))
                # node = node.copy(name)
                logger.warning(
                    f"{node} has to be copied due to conflicting name ({name})"
                )
            new_nodes[name] = node
        return new_nodes

    # TODO: logic commonly same with node so need to move to
    @classmethod
    def scan_folder_for_instances(cls, path: Path) -> dict[str, QGraphBaseType]:
        """
        Scans a folder for graph instances and returns them.

        This method scans the specified folder for graph files that represent
        valid instances of `QualibrationGraph`.

        Args:
            path (Path): The folder to scan for graph files.

        Returns:
            dict[str, QGraphBaseType]: A dictionary of graph names to their
                corresponding instances.
        """
        graphs: dict[str, QGraphBaseType] = {}
        try:
            if run_modes_ctx.get() is not None:
                logger.error(
                    "Run modes context is already set to %s",
                    run_modes_ctx.get(),
                )
            run_modes_token = run_modes_ctx.set(RunModes(inspection=True))

            for file in sorted(path.iterdir()):
                if not file_is_calibration_instance(file, cls.__name__):
                    continue
                try:
                    cls.scan_graph_file(file, graphs)
                except Exception as e:
                    logger.exception("", exc_info=e)
                    logger.warning(
                        "An error occurred on scanning graph file "
                        f"{file.name}.\nError message: {e}"
                    )
        finally:
            run_modes_ctx.reset(run_modes_token)
        return graphs

    @classmethod
    def scan_graph_file(
        cls, file: Path, graphs: dict[str, QGraphBaseType]
    ) -> None:
        """
        Scans a graph file and adds its instance to the given dictionary.

        This method inspects the content of a given graph file and adds a valid
        `QualibrationGraph` to the provided dictionary.

        Args:
            file (Path): The file to scan for a graph.
            graphs (dict[str, QGraphBaseType]): The dictionary to add valid
                graphs to.

        Raises:
            StopInspection: Used to stop execution once inspection completes.
        """
        logger.info(f"Scanning graph file {file}")
        try:
            # TODO Think of a safer way to execute the code
            _module = import_from_path(get_module_name(file), file)
        except StopInspection as ex:
            graph = cast("QualibrationGraph[NodeTypeVar]", ex.instance)
            graph.filepath = file
            graph.modes.inspection = False
            cls.add_graph(graph, graphs)

    @classmethod
    def add_graph(
        cls,
        graph: "QualibrationGraph[NodeTypeVar]",
        graphs: dict[str, QGraphBaseType],
    ) -> None:
        """
        Adds a graph to the library dictionary.

        If a graph with the same name already exists, it overwrites it with
        a warning.

        Args:
            graph (QualibrationGraph): The graph to add.
            graphs (dict[str, QGraphBaseType]): The dictionary to store the
                graph.
        """
        if graph.name in graphs:
            logger.warning(
                f'Graph "{graph.name}" already exists in library, overwriting'
            )

        graphs[graph.name] = graph

    def cleanup(self) -> None:
        """
        Cleans up the graph and resets nodes to their initial states.

        This method is used to clear any progress made during execution by
        resetting nodes to their initial states.
        """
        super().cleanup()
        nx.set_node_attributes(
            self._graph,
            {node: self._node_init_args.copy() for node in self._graph.nodes},
        )
        if self._orchestrator:
            self._orchestrator.cleanup()

    def completed_count(self) -> int:
        """
        Returns the count of completed nodes in the graph.

        This method calculates how many nodes in the graph have been completed
        (i.e., they are no longer pending).

        Returns:
            int: The count of nodes that have been completed.
        """
        return int(
            sum(
                map(
                    lambda status: status != NodeStatus.pending,
                    nx.get_node_attributes(
                        self._graph, self.STATUS_FIELD
                    ).values(),
                )
            )
        )

    @property
    def active_node_name(self) -> Optional[str]:
        return (
            self._orchestrator.active_node_name
            if self._orchestrator is not None
            else None
        )

    def _get_all_nodes_parameters(
        self, nodes_parameters: Mapping[str, Any]
    ) -> Mapping[str, Any]:
        """
        Retrieves parameters for all nodes, providing defaults if unspecified.

        This method extracts the parameters for each node in the graph.
        If parameters are missing for any node, it assigns default values.

        Args:
            nodes_parameters (Mapping[str, Any]): dictionary containing
                parameters for nodes, keyed by node names.

        Returns:
            Mapping[str, Any]: A dictionary containing parameters for all nodes,
            ensuring that each node has a corresponding entry.
        """
        nodes_class = self.full_parameters_class.model_fields[
            "nodes"
        ].annotation
        return {
            name: nodes_parameters.get(name, {})
            for name in cast(NodesParameters, nodes_class).model_fields
        }

    def _orchestrator_or_error(self) -> "QualibrationOrchestrator[NodeTypeVar]":
        """
        Retrieves the orchestrator for the graph or raises an error if missing.

        This method returns the orchestrator associated with the graph.
        If no orchestrator is specified, it raises an error indicating
        that an orchestrator is required for execution.

        Returns:
            QualibrationOrchestrator: The orchestrator used to manage graph
                execution.

        Raises:
            ValueError: If no orchestrator is specified for the graph.
        """
        if self._orchestrator is None:
            ex = ValueError("Orchestrator not specified")
            logger.exception("", exc_info=ex)
            raise ex
        return self._orchestrator

    def _run(self, **passed_parameters: Any) -> None:
        """
        Runs the graph by traversing nodes using the orchestrator.

        This method performs the actual execution of the graph by validating
        the passed parameters and invoking the orchestrator to traverse through
        the graph.

        Args:
            **passed_parameters (Any): Parameters passed for graph execution.
        """
        orchestrator = self._orchestrator_or_error()
        self.cleanup()
        nodes = self._get_all_nodes_parameters(
            passed_parameters.get("nodes", {})
        )
        self._parameters = self.parameters.model_validate(passed_parameters)
        self.full_parameters = self.full_parameters_class.model_validate(
            {"parameters": self.parameters, "nodes": nodes}
        )
        targets = self._initial_targets = (
            self.full_parameters.parameters.targets or []
        )
        nodes_parameters_model = self.full_parameters.nodes
        for node_name in nodes_parameters_model.model_fields_set:
            node_parameters_model = getattr(nodes_parameters_model, node_name)
            if node_parameters_model.targets_name is not None:
                node_parameters_model.targets = targets
        orchestrator.traverse_graph(self, targets)

    def _post_run(
        self,
        created_at: datetime,
        run_error: Optional[RunError],
    ) -> GraphRunSummary:
        """
        Finalizes the graph execution and generates a summary.

        This method updates the outcomes of the graph, creates a summary
        with information about the execution, including successful and failed
        targets, and logs the summary.

        Args:
            created_at (datetime): Timestamp when the graph run was started.
            run_error (Optional[RunError]): Details of any error encountered
                during execution.

        Returns:
            GraphRunSummary: A summary object containing details about the
                graph run.
        """
        self.outcomes = self._orchestrator_or_error().final_outcomes
        self.run_summary = GraphRunSummary(
            name=self.name,
            description=self.description,
            created_at=created_at,
            completed_at=datetime.now(),
            parameters=self.full_parameters,
            outcomes=self.outcomes,
            initial_targets=self._initial_targets,
            error=run_error,
            successful_targets=[
                name
                for name, status in self.outcomes.items()
                if status == Outcome.SUCCESSFUL
            ],
            failed_targets=[
                name
                for name, status in self.outcomes.items()
                if status == Outcome.FAILED
            ],
            state_updates=self.state_updates,
        )
        logger.debug(f"Graph run summary {self.run_summary}")
        return self.run_summary

    def run(
        self, **passed_parameters: Any
    ) -> tuple["QualibrationGraph[NodeTypeVar]", BaseRunSummary]:
        """
        Runs the graph using the given parameters.

        This method orchestrates the execution of all nodes in the graph,
        following the specified connectivity and using the provided parameters.

        Args:
            **passed_parameters (Any): Graph parameters to use for the
                execution. Should include the `nodes` key.

        Returns:
            tuple[QualibrationGraph, BaseRunSummary]: The graph and a summary of
            the execution, including outcomes and other details.

        Raises:
            RunError: If any error occurs during graph execution.
        """
        logger.info(
            f"Run graph {self.name} with parameters: {passed_parameters}"
        )
        created_at = datetime.now()
        run_error: Optional[RunError] = None
        try:
            self._run(**passed_parameters)
        except Exception as ex:
            run_error = RunError(
                error_class=ex.__class__.__name__,
                message=str(ex),
                traceback=traceback.format_tb(ex.__traceback__),
            )
            raise
        finally:
            run_summary = self._post_run(created_at, run_error)
        return self, run_summary

    def _get_qnode_or_error(self, node_name: str) -> NodeTypeVar:
        """
        Retrieves a node by name or raises an error if the node is not found.

        This method fetches a node from the graph using the given name.
        If the node is not found in the graph, an error is raised.

        Args:
            node_name (str): The name of the node to retrieve.

        Returns:
            QualibrationNode: The node associated with the given name.

        Raises:
            ValueError: If no node with the specified name exists.
        """
        # node = cast(NodeType, self._nodes.get(node_name))
        node = self._nodes.get(node_name)
        if node is None:
            raise ValueError(f"Unknown node with name {node_name}", node_name)
        return node

    def _add_node_by_name(self, node_name: str) -> NodeTypeVar:
        """
        Adds a node to the graph by name, if not already present.

        This method adds the node identified by `node_name` to the graph,
        including initializing it with default attributes.

        Args:
            node_name (str): The name of the node to add.

        Returns:
            QualibrationNode: The node that was added or already exists in the
                graph.
        """
        node = self._get_qnode_or_error(node_name)
        if node not in self._graph:
            self._graph.add_node(node, **self.__class__._node_init_args)
        return node

    def _build_parameters_class(self) -> type[GraphRunParametersType]:
        """
        Builds a class for the parameters used in running the graph.

        This method dynamically creates a model class that combines the
        parameters for the graph itself and its nodes, to be used during the
        graph's execution.

        Returns:
            The parameter class to be used for
                execution.
        """
        nodes_parameters_class = create_model(
            "GraphNodesParameters",
            __base__=NodesParameters,
            **{  # type: ignore
                node.name: (node.parameters_class, node.parameters)
                for node in self._graph.nodes
            },
        )
        execution_parameters_class = create_model(
            "ExecutionParameters",
            __base__=ExecutionParameters,
            parameters=(self.parameters.__class__, self.parameters),
            nodes=(nodes_parameters_class, nodes_parameters_class()),
        )
        return execution_parameters_class

    def serialize(self, **kwargs: Any) -> Mapping[str, Any]:
        """
        Serializes the graph into a dictionary format.

        This method provides a JSON-serializable representation of the graph,
        including nodes, parameters, connectivity, orchestrator.

        Args:
            **kwargs (Any): Additional arguments for serialization.

        Keyword Args:
            cytoscape (bool): Is it needed to include cytoscape information.
                Defaults to `False`.
        Returns:
            Mapping[str, Any]: Serialized representation of the graph.
        """
        data = dict(super().serialize())
        cytoscape = bool(kwargs.get("cytoscape", False))
        parameters = self.full_parameters_class.serialize(**kwargs)
        nx_data: dict[str, Any] = dict(
            self.nx_graph_export(node_names_only=True)
        )
        data.update(
            {
                "parameters": parameters["parameters"],
                "orchestrator": (
                    self._orchestrator.serialize()
                    if self._orchestrator is not None
                    else None
                ),
            }
        )
        nodes = {}
        connectivity = []
        for node, adjacency in zip(
            nx_data.pop("nodes"), nx_data.pop("adjacency")
        ):
            node_id = node["id"]
            nodes[node_id] = node
            node.update(
                {
                    # TODO: simplify node name
                    "name": node_id,
                    "parameters": parameters["nodes"][node["id"]],
                }
            )
            connectivity.extend([(node_id, item["id"]) for item in adjacency])
        data.update({"nodes": nodes, "connectivity": connectivity})
        if cytoscape:
            data["cytoscape"] = self.cytoscape_representation(data)
        return data

    def nx_graph_export(
        self, node_names_only: bool = False
    ) -> Mapping[str, Any]:
        """
        Exports the graph as a networkx adjacency list.

        Args:
            node_names_only (bool): If True, only node names are included in
                the adjacency list. Defaults to False.

        Returns:
            Mapping[str, Any]: A dictionary representing the adjacency list of
            the graph.
        """
        data = dict(nx.readwrite.adjacency_data(self._graph))
        for key in ("multigraph", "directed", "graph"):
            data.pop(key)
        if node_names_only:
            for node, adjacency in zip(data["nodes"], data["adjacency"]):
                node["id"] = node["id"].name
                for adj in adjacency:
                    adj["id"] = adj["id"].name
        return data

    def cytoscape_representation(
        self, serialized: Mapping[str, Any]
    ) -> Sequence[Mapping[str, Any]]:
        """
        Returns a Cytoscape-compatible representation of the graph.

        This method generates nodes and edges in a format that can be used with
        Cytoscape for visualization purposes.

        Args:
            serialized (Mapping[str, Any]): Serialized representation of the
                graph.

        Returns:
            Sequence[Mapping[str, Any]]: List of nodes and edges for Cytoscape
                visualization.
        """
        nodes = [
            {
                "group": "nodes",
                "data": {"id": node},
                "position": {"x": 100, "y": 100},
            }
            for node in serialized["nodes"]
        ]
        edges = [
            {
                "group": "edges",
                "data": {
                    "id": f"{source}_{dest}",
                    "source": source,
                    "target": dest,
                },
            }
            for source, dest in serialized["connectivity"]
        ]
        return [*nodes, *edges]

    def stop(self, **kwargs: Any) -> bool:
        """
        Stops the execution of the graph or nodes within it.

        This method stops the orchestrator associated with the graph and
        optionally stops any active nodes.

        Args:
            **kwargs (Any): Additional arguments.

        Keyword Args:
            stop_graph_node (bool): Is it needed to stop running node.

        Returns:
            bool: True if successful in stopping execution, False otherwise.
        """
        logger.debug(f"Stop graph {self.name}")
        stop_node: Optional[bool] = kwargs.get("stop_graph_node")
        node_stop = True
        orchestrator = self._orchestrator
        if orchestrator is None:
            return False
        if stop_node and (node := orchestrator.active_node):
            node_stop = node.stop()
        orchestrator.stop()
        return node_stop



from pathlib import Path
from typing import Any, Union

import numpy as np
from numpy.lib.npyio import NpzFile

from qualibrate.utils.node.loaders.base_loader import BaseLoader


class NumpyArrayLoader(BaseLoader):
    """
    Loader for NumPy array files.

    Attributes:
        file_extensions: A tuple of supported file extensions
            (e.g., ".npy", ".npz").
        filepath_to_array: A cache mapping file paths to loaded NumPy arrays.
    """

    file_extensions = (".npy", ".npz")

    def __init__(self) -> None:
        self.filepath_to_array: dict[
            Path, Union[np.ndarray[Any, np.dtype[Any]], NpzFile]
        ] = {}

    def load(self, file_path: Path, **kwargs: Any) -> Any:
        """
        Loads a NumPy array file and resolves subreferences if applicable.

        Args:
            file_path: The path to the NumPy file.
            **kwargs: Additional arguments, including "subref" for subreference
                keys.

        Returns:
            The loaded NumPy array or the specified subreference content.

        Raises:
            ValueError: If the file is not an `NpzFile` but a subreference is
                requested.
        """
        if file_path in self.filepath_to_array:
            file_content = self.filepath_to_array[file_path]
        else:
            self.__class__.validate_file_exists(file_path)
            file_content = np.load(file_path)
            self.filepath_to_array[file_path] = file_content
        subref = kwargs.get("subref")
        if subref is None:
            return file_content
        if not isinstance(file_content, NpzFile):
            raise ValueError(
                f"Loaded file {file_path} is not representation of "
                f"multiple NumPy arrays"
            )
        return file_content.get(subref)



from collections import UserDict, UserList
from typing import TYPE_CHECKING, Any, Optional, TypeVar

from qualibrate.parameters import NodeParameters
from qualibrate.utils.type_protocols import (
    GetRefGetItemProtocol,
    GetRefProtocol,
)

if TYPE_CHECKING:
    from qualibrate.qualibration_node import QualibrationNode


__all__ = [
    "record_state_update_getattr",
    "record_state_update_getitem",
]

ParametersType = TypeVar("ParametersType", bound=NodeParameters)


def _record_state_update(
    node: Optional["QualibrationNode[ParametersType]"],
    reference: str,
    attr: str,
    old: Any,
    val: Any,
) -> None:
    """
    Records state updates for an attribute or item in the node.

    This function stores information about changes made to an attribute
    or item of a node, including the previous value and the new value.
    If the node is provided, the change details are saved in the node's
    `_state_updates` dictionary.

    Args:
        node: The node where the state update will be recorded. If None, no
            action is performed.
        reference: The reference key to identify the updated attribute or item.
        attr: The name of the attribute or item key that is updated.
        old: The old value of the attribute or item before the update.
        val: The new value of the attribute or item.
    """
    if node is None:
        return
    if isinstance(old, UserList):
        old = list(old)
    elif isinstance(old, UserDict):
        old = dict(old)
    node._state_updates[reference] = {
        "key": reference,
        "attr": attr,
        "old": old,
        "new": val,
    }


def record_state_update_getattr(
    quam_obj: GetRefProtocol,
    attr: str,
    val: Any = None,
    node: Optional["QualibrationNode[ParametersType]"] = None,
) -> None:
    """
    Records item state updates in a Quam collection object.

    For details see `_record_state_update`.

    Args:
        quam_obj: The Quam object whose attribute is updated.
        attr: The name of the attribute being updated.
        val: The new value of the attribute. Defaults to None.
        node: The node where the state update will be recorded. Defaults to
            None.
    """
    _record_state_update(
        node, quam_obj.get_reference(attr), attr, getattr(quam_obj, attr), val
    )


def record_state_update_getitem(
    quam_obj: GetRefGetItemProtocol,
    attr: str,
    val: Any = None,
    node: Optional["QualibrationNode[ParametersType]"] = None,
) -> None:
    """
    Records item state updates in a Quam collection object.

    For details see `_record_state_update`.

    Args:
        quam_obj: The Quam object whose item is being updated.
        attr: The key/index of the item being updated.
        val: The new value of the item. Defaults to None.
        node: The node where the state update will be recorded. Defaults to
            None.
    """
    _record_state_update(
        node, quam_obj.get_reference(attr), attr, quam_obj[attr], val
    )



import traceback
from collections.abc import Sequence
from datetime import datetime
from queue import Queue
from typing import Any, Generic, Optional

import networkx as nx

from qualibrate import QualibrationGraph
from qualibrate.models.execution_history import ExecutionHistoryItem
from qualibrate.models.node_status import NodeStatus
from qualibrate.models.outcome import Outcome
from qualibrate.models.run_summary.run_error import RunError
from qualibrate.orchestration.qualibration_orchestrator import (
    QualibrationOrchestrator,
)
from qualibrate.qualibration_graph import NodeTypeVar
from qualibrate.utils.logger_m import logger

__all__ = ["BasicOrchestrator"]


class BasicOrchestrator(
    QualibrationOrchestrator[NodeTypeVar], Generic[NodeTypeVar]
):
    """
    A basic orchestrator that manages the execution of nodes in a graph.
    This orchestrator firstly run nodes without predecessors. And then just
    walk through directed graph for resolving nodes that can be executed
    (all predecessors are completed).

    Args:
        skip_failed (bool): If True, when a target (qubit) fails in a node,
            it is excluded from being a target for any future nodes.

    """

    def __init__(self, skip_failed: bool = False):
        """
        Initializes a new instance of the BasicOrchestrator.

        """
        super().__init__(skip_failed=skip_failed)
        self._execution_queue: Queue[NodeTypeVar] = Queue()

    def _is_execution_finished(self) -> bool:
        """
        Checks whether the execution is finished.

        Returns:
            bool: True if the execution is finished, False otherwise.
        """
        if self._graph is None:
            return True
        if self._execution_queue.qsize() == 0:
            return True
        return all(
            map(
                lambda status: status != NodeStatus.pending,
                nx.get_node_attributes(
                    self.nx_graph, QualibrationGraph.STATUS_FIELD
                ).values(),
            )
        )

    def cleanup(self) -> None:
        """
        Cleans up the orchestrator state.

        Clears the execution queue and calls the parent cleanup method.
        """
        super().cleanup()
        with self._execution_queue.mutex:
            self._execution_queue.queue.clear()

    @property
    def nx_graph(self) -> "nx.DiGraph[NodeTypeVar]":
        """
        Gets the networkx representation of the graph.

        Returns:
            nx.DiGraph[QualibrationNode]: The directed graph.

        Raises:
            ValueError: If the graph is not specified.
        """
        if self._graph is None:
            raise ValueError("Graph is not specified")
        return self._graph._graph

    def check_node_successful(self, node: NodeTypeVar) -> bool:
        """
        Checks if a node was successfully executed.

        Args:
            node (QualibrationNode): The node to check.

        Returns:
            bool: True if the node is successful, False otherwise.
        """
        if self._graph is None:
            return False
        return bool(
            self.nx_graph.nodes[node][QualibrationGraph.STATUS_FIELD]
            == NodeStatus.successful
        )

    def get_next_node(self) -> Optional[NodeTypeVar]:
        """
        Gets the next node to execute.

        Returns:
            Optional[QualibrationNode]: The next node to execute, or None
                if there are no more nodes to be executed
        """
        while not self._execution_queue.empty():
            node_to_run = self._execution_queue.get()
            if all(
                map(self.check_node_successful, self.nx_graph.pred[node_to_run])
            ):
                return node_to_run
        return None

    def traverse_graph(
        self, graph: QualibrationGraph[NodeTypeVar], targets: Sequence[Any]
    ) -> None:
        """
        Traverses the graph and orchestrates node execution.

        Args:
            graph (QualibrationGraph): The graph to traverse.
            targets (Sequence[Any]): The target nodes to execute.

        Raises:
            RuntimeError:
                If graph parameters are not specified or no next node is found.
        """
        logger.info(f"Traverse graph {graph.name} with targets {targets}")
        if self._is_stopped:
            return
        self._graph = graph
        if graph.full_parameters is None:
            ex = RuntimeError("Execution graph parameters not specified")
            logger.exception("", exc_info=ex)
            raise ex
        self.initial_targets = list(targets[:]) if targets else []
        self.targets = self.initial_targets.copy()
        nodes_parameters = graph.full_parameters.nodes
        nx_graph = self.nx_graph
        predecessors = nx_graph.pred
        successors = nx_graph.succ
        nodes_without_predecessors = filter(
            lambda n: len(predecessors[n]) == 0, predecessors.keys()
        )
        for node in nodes_without_predecessors:
            self._execution_queue.put(node)

        while not self._is_execution_finished() and not self._is_stopped:
            node_to_run = self.get_next_node()
            logger.info(f"Graph. Node to run. {node_to_run}")
            if node_to_run is None:
                exc = RuntimeError("No next node. Execution not finished")
                logger.exception("", exc_info=exc)
                raise exc
            node_to_run_parameters = getattr(nodes_parameters, node_to_run.name)
            run_start = datetime.now()
            run_error: Optional[RunError] = None
            try:
                self._active_node = node_to_run
                node_to_run_parameters.targets = self.targets
                node_parameters = node_to_run_parameters.model_dump()
                logger.debug(
                    f"Graph. Start running node {node_to_run} "
                    f"with parameters {node_parameters}"
                )
                executed_node, node_result = node_to_run.run(
                    interactive=False, **node_parameters
                )
                if self._parameters.skip_failed:
                    self.targets = node_result.successful_targets
                logger.debug(f"Node completed. Result: {node_result}")
            except Exception as ex:
                new_status = NodeStatus.failed
                executed_node = node_to_run
                nx_graph.nodes[node_to_run]["error"] = str(ex)
                logger.exception(
                    (
                        f"Failed to run node {node_to_run.name} "
                        f"in graph {self._graph.name}"
                    ),
                    exc_info=ex,
                )
                run_error = RunError(
                    error_class=ex.__class__.__name__,
                    message=str(ex),
                    traceback=traceback.format_tb(ex.__traceback__),
                )
                raise
            else:
                new_status = NodeStatus.successful
            finally:
                self._execution_history.append(
                    ExecutionHistoryItem(
                        name=executed_node.name,
                        description=executed_node.description,
                        snapshot_idx=executed_node.snapshot_idx,
                        outcomes=executed_node.outcomes,
                        status=new_status,
                        error=run_error,
                        run_start=run_start,
                        run_end=datetime.now(),
                        parameters=executed_node._parameters,
                    )
                )
            # Suppose that all nodes are successfully finish
            nx_graph.nodes[node_to_run][QualibrationGraph.STATUS_FIELD] = (
                new_status
            )
            if new_status == NodeStatus.successful:
                for successor in successors[node_to_run]:
                    self._execution_queue.put(successor)
        self._active_node = None
        # TODO: correct resolving of outcomes
        for target in targets:
            self.final_outcomes[target] = Outcome.SUCCESSFUL



import logging
from collections.abc import Mapping
from logging.handlers import RotatingFileHandler
from pathlib import Path
from types import TracebackType
from typing import Optional, Union

from qualibrate_config.resolvers import (
    get_qualibrate_config,
    get_qualibrate_config_path,
)

_SysExcInfoType = Union[
    tuple[type[BaseException], BaseException, Optional[TracebackType]],
    tuple[None, None, None],
]
_ExcInfoType = Optional[Union[bool, _SysExcInfoType, BaseException]]
_ArgsType = Union[tuple[object, ...], Mapping[str, object]]

__all__ = ["logger"]

LOG_FORMAT = (
    "%(asctime)s - %(name)s - %(levelname)s - %(message)s "
    "(%(filename)s:%(lineno)d)"
)


class QualibrateFormatter(logging.Formatter):
    formatter = logging.Formatter(LOG_FORMAT)

    def format(self, record: logging.LogRecord) -> str:
        return self.formatter.format(record)


class ConsoleFormatter(QualibrateFormatter):
    grey = "\x1b[38;20m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"

    FORMATS = {
        logging.DEBUG: grey + LOG_FORMAT + reset,
        logging.INFO: grey + LOG_FORMAT + reset,
        logging.WARNING: yellow + LOG_FORMAT + reset,
        logging.ERROR: red + LOG_FORMAT + reset,
        logging.CRITICAL: bold_red + LOG_FORMAT + reset,
    }
    FORMATTERS = {
        level: logging.Formatter(format) for level, format in FORMATS.items()
    }

    def format(self, record: logging.LogRecord) -> str:
        log_fmtr = self.FORMATTERS.get(record.levelno, self.formatter)
        return log_fmtr.format(record)


class LazyInitLogger(logging.Logger):
    _initialized = False

    def _log(
        self,
        level: int,
        msg: object,
        args: _ArgsType,
        exc_info: Optional[_ExcInfoType] = None,
        extra: Optional[Mapping[str, object]] = None,
        stack_info: bool = False,
        stacklevel: int = 1,
    ) -> None:
        if not self._initialized:
            self._initialize()
        super()._log(level, msg, args, exc_info, extra, stack_info)

    @staticmethod
    def get_log_filepath() -> Path:
        q_config_path = get_qualibrate_config_path()
        qs = get_qualibrate_config(q_config_path)
        log_folder = qs.log_folder
        if log_folder is None:
            log_folder = Path().home().joinpath(".qualibrate", "logs")
        log_folder.mkdir(parents=True, exist_ok=True)
        log_file_path = log_folder / "qualibrate.log"
        return log_file_path

    def _initialize(self) -> None:
        self.setLevel(logging.DEBUG)
        file_handler = RotatingFileHandler(
            self.get_log_filepath(),
            maxBytes=1024 * 1024 * 10,  # 10 Mb
            backupCount=3,
        )
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(QualibrateFormatter())
        self.addHandler(file_handler)

        console = logging.StreamHandler()
        console.setLevel(logging.INFO)
        console.setFormatter(ConsoleFormatter())
        self.addHandler(console)
        self._initialized = True


logger = LazyInitLogger("qualibrate-core")



import json
from pathlib import Path
from typing import TYPE_CHECKING, Generic, Optional, TypeVar

from qualang_tools.results import DataHandler

from qualibrate.storage.storage_manager import StorageManager
from qualibrate.utils.logger_m import logger

if TYPE_CHECKING:
    from qualibrate.parameters import NodeParameters
    from qualibrate.qualibration_node import QualibrationNode


NodeTypeVar = TypeVar("NodeTypeVar", bound="QualibrationNode[NodeParameters]")


class LocalStorageManager(StorageManager[NodeTypeVar], Generic[NodeTypeVar]):
    """
    Manages local storage of calibration nodes and their machine states.

    The `LocalStorageManager` class is responsible for saving the state of
    `QualibrationNode` instances to a local file system. It extends the
    `StorageManager` abstract base class, providing a concrete implementation
    that uses a specified root folder for data storage.

    Args:
        root_data_folder (Path): The root folder where data should be saved.
        active_machine_path (Optional[Path]): Optional path for saving the
            current active machine state.
    """

    machine_content_mapping = {"wiring.json": {"wiring", "network"}}

    def __init__(
        self, root_data_folder: Path, active_machine_path: Optional[Path] = None
    ):
        self.root_data_folder = root_data_folder
        self.data_handler = DataHandler(root_data_folder=root_data_folder)
        self.active_machine_path = active_machine_path
        self.snapshot_idx = None

    def save(self, node: NodeTypeVar) -> None:
        """
        Saves the state of the specified node to local storage.

        This method saves the results of the node to the root data folder.
        It also saves the machine state, either as a JSON file or as a folder
        with specific content mappings, depending on the structure of the
        machine. Optionally, it can also save the active machine state to a
        specified path.

        Args:
            node (QualibrationNode): The node whose state needs to be saved.

        Side Effects:
            - Saves the node's results and machine state in the root data
                folder.
            - Updates the `snapshot_idx` to reflect the newly saved node state.
            - Optionally saves the machine state to the active path if
                specified.

        Raises:
            AssertionError: If `self.data_handler.path` is not of type `Path`.
        """
        logger.info(f"Saving node {node.name} to local storage")

        # Save results
        self.data_handler.name = node.name
        DataHandler.node_data = {
            "quam": "./quam_state.json",
            "parameters": {
                "model": node.parameters.model_dump(mode="json"),
                "schema": node.parameters.__class__.model_json_schema(),
            },
        }
        node_contents = (
            self.data_handler.generate_node_contents()
        )  # TODO directly access idx
        self.data_handler.save_data(
            data=node.results,
            name=node.name,
            node_contents=node_contents,
        )
        self.snapshot_idx = node_contents["id"]

        if node.machine is None:
            logger.info("Node has no QuAM, skipping machine.save")
            return

        assert isinstance(self.data_handler.path, Path)  # TODO Remove assertion

        if self.machine_content_mapping is None:
            content_mapping = None
        elif all(
            hasattr(node.machine, elem)
            for elem_group in self.machine_content_mapping.values()
            for elem in elem_group
        ):
            content_mapping = self.machine_content_mapping
        else:
            content_mapping = None

        # Save QuAM to the data folder
        if isinstance(node.machine, dict):
            quam_path = self.data_handler.path / "quam_state2.json"
            quam_path.write_text(
                json.dumps(node.machine, indent=4, sort_keys=True)
            )
        else:
            # Save as single file
            node.machine.save(path=self.data_handler.path / "quam_state.json")
            # Save as folder with wiring and network separated
            if content_mapping is not None:
                node.machine.save(
                    path=self.data_handler.path / "quam_state",
                    content_mapping=content_mapping,
                )

        # Optionally also save QuAM to the active path
        if self.active_machine_path is not None:
            logger.info(
                f"Saving machine to active path {self.active_machine_path}"
            )
            node.machine.save(
                path=self.active_machine_path,
                content_mapping=content_mapping,
            )



from pathlib import Path
from typing import Any

from qualibrate.utils.logger_m import logger
from qualibrate.utils.node.loaders.json_loader import JSONLoader

try:
    from quam.core import QuamRoot
except ModuleNotFoundError:
    QuamRoot = None


class QuamLoader(JSONLoader):
    """Loader for QUAM files, extending JSONLoader."""

    def load(self, file_path: Path, **kwargs: Any) -> Any:
        """
        Loads a QUAM file and attempts to parse it using `QuamRoot`.

        Args:
            file_path: The path to the QUAM file.
            **kwargs: Additional arguments for file loading.

        Returns:
            The parsed QUAM object, or the raw JSON content if parsing fails.
        """
        json_data = super().load(file_path, **kwargs)
        if QuamRoot is None:
            return json_data
        try:
            return QuamRoot.load(json_data)
        except Exception as ex:
            logger.exception(
                "QuamRoot wasn't successfully loaded from reference "
                f"{file_path}. Returning raw json",
                exc_info=ex,
            )
            return json_data



from typing import Any, Optional, Protocol

TargetType = str


class GetRefProtocol(Protocol):
    def get_reference(self, attr: Optional[str]) -> str: ...


class GetRefGetItemProtocol(GetRefProtocol, Protocol):
    def __getitem__(self, attr: Optional[str]) -> Any:
        pass



from collections.abc import Mapping, Sequence
from datetime import datetime
from typing import Any, Optional

from pydantic import BaseModel, Field, computed_field, field_serializer

from qualibrate.models.outcome import Outcome
from qualibrate.models.run_summary.run_error import RunError
from qualibrate.parameters import RunnableParameters
from qualibrate.utils.type_protocols import TargetType

__all__ = ["BaseRunSummary"]


class BaseRunSummary(BaseModel):
    name: str
    description: Optional[str] = None
    created_at: datetime
    completed_at: datetime
    parameters: Optional[RunnableParameters] = None
    outcomes: dict[TargetType, Outcome]
    error: Optional[RunError] = None

    initial_targets: Sequence[TargetType] = Field(default_factory=list)
    successful_targets: list[TargetType] = Field(default_factory=list)
    failed_targets: list[TargetType] = Field(default_factory=list)
    dropped_targets: Optional[list[TargetType]] = None
    state_updates: Mapping[str, Any] = Field(default_factory=dict)

    @computed_field
    def run_duration(self) -> float:
        return round((self.completed_at - self.created_at).total_seconds(), 3)

    @field_serializer("parameters")
    def serialize_parameters(
        self, parameters: RunnableParameters
    ) -> Mapping[str, Any]:
        return parameters.model_dump(serialize_as_any=True)



from collections.abc import Sequence
from datetime import datetime
from typing import Optional

from pydantic import BaseModel, ConfigDict, Field, computed_field

from qualibrate.models.node_status import NodeStatus
from qualibrate.models.outcome import Outcome
from qualibrate.models.run_summary.run_error import RunError
from qualibrate.parameters import NodeParameters
from qualibrate.utils.type_protocols import TargetType

__all__ = ["ExecutionHistory", "ExecutionHistoryItem"]


class ExecutionHistoryItem(BaseModel):
    """Represents an item of graph execution history."""

    model_config = ConfigDict()

    name: str
    description: Optional[str] = None
    snapshot_idx: Optional[int] = None
    status: NodeStatus
    run_start: datetime
    run_end: datetime
    parameters: NodeParameters
    error: Optional[RunError] = None
    outcomes: dict[TargetType, Outcome] = Field(default_factory=dict)

    @computed_field
    def run_duration(self) -> float:
        """Time in seconds node run"""
        return round((self.run_end - self.run_start).total_seconds(), 3)


class ExecutionHistory(BaseModel):
    """Represents a graph execution history."""

    items: Sequence[ExecutionHistoryItem]



from typing import Optional

from qualibrate import NodeParameters
from qualibrate.models.run_summary.base import BaseRunSummary

__all__ = ["NodeRunSummary"]


class NodeRunSummary(BaseRunSummary):
    snapshot_idx: Optional[int] = None
    parameters: NodeParameters



import traceback
from collections.abc import Generator, Mapping, Sequence
from contextlib import contextmanager
from contextvars import ContextVar
from copy import copy
from datetime import datetime
from functools import partialmethod
from importlib.util import find_spec
from pathlib import Path
from typing import (
    Any,
    Optional,
    TypeVar,
    Union,
    cast,
)

import matplotlib
from matplotlib.backends import (  # type: ignore[attr-defined]
    BackendFilter,
    backend_registry,
)
from pydantic import ValidationError, create_model
from qualibrate_config.resolvers import (
    get_qualibrate_config,
    get_qualibrate_config_path,
)

from qualibrate.config.resolvers import get_quam_state_path
from qualibrate.models.outcome import Outcome
from qualibrate.models.run_mode import RunModes
from qualibrate.models.run_summary.base import BaseRunSummary
from qualibrate.models.run_summary.node import NodeRunSummary
from qualibrate.models.run_summary.run_error import RunError
from qualibrate.parameters import NodeParameters
from qualibrate.q_runnnable import (
    QRunnable,
    file_is_calibration_instance,
    run_modes_ctx,
)
from qualibrate.storage import StorageManager
from qualibrate.storage.local_storage_manager import LocalStorageManager
from qualibrate.utils.exceptions import StopInspection
from qualibrate.utils.logger_m import logger
from qualibrate.utils.node.comined_method import InstanceOrClassMethod
from qualibrate.utils.node.content import (
    parse_node_content,
    read_node_content,
    read_node_data,
)
from qualibrate.utils.node.loaders.base_loader import BaseLoader
from qualibrate.utils.node.path_solver import (
    get_node_dir_path,
)
from qualibrate.utils.node.record_state_update import (
    record_state_update_getattr,
    record_state_update_getitem,
)
from qualibrate.utils.read_files import get_module_name, import_from_path
from qualibrate.utils.type_protocols import TargetType

__all__ = [
    "QualibrationNode",
    "NodeCreateParametersType",
    "NodeRunParametersType",
]

NodeCreateParametersType = NodeParameters
NodeRunParametersType = NodeParameters
ParametersType = TypeVar("ParametersType", bound=NodeParameters)

# TODO: use node parameters type instead of Any
external_parameters_ctx: ContextVar[Optional[tuple[str, Any]]] = ContextVar(
    "external_parameters", default=None
)
last_executed_node_ctx: ContextVar[Optional["QualibrationNode[Any]"]] = (
    ContextVar("last_executed_node", default=None)
)


class QualibrationNode(
    QRunnable[ParametersType, NodeRunParametersType],
):
    """
    Represents a qualibration node (that can be run independently or as part
    of graph), responsible for executing specific tasks with defined parameters
     and modes.

    Args:
        name: The name of the node.
        parameters: Parameters passed to the node for its initialization.
            Defaults to None.
        description: A description of the node. Defaults to None.
        parameters_class: Class used for node parameters validation. Defaults
            to None.
        modes: Execution modes. Defaults to None.

    Raises:
        StopInspection: Raised if the node is instantiated in inspection mode.
    """

    storage_manager: Optional[
        StorageManager["QualibrationNode[NodeParameters]"]
    ] = None
    active_node: Optional["QualibrationNode[ParametersType]"] = None

    def __init__(
        self,
        name: str,
        parameters: Optional[ParametersType] = None,
        description: Optional[str] = None,
        *,
        parameters_class: Optional[type[ParametersType]] = None,
        modes: Optional[RunModes] = None,
    ):
        logger.info(f"Creating node {name}")
        parameters = self.__class__._validate_passed_parameters_options(
            name, parameters, parameters_class
        )
        super().__init__(
            name,
            parameters,
            description=description,
            modes=modes,
        )

        self.results: dict[Any, Any] = {}
        self.machine = None

        if self.modes.inspection:
            raise StopInspection(
                "Node instantiated in inspection mode", instance=self
            )
        self.__class__.active_node = self
        last_executed_node_ctx.set(self)

        self._warn_if_external_and_interactive_mpl()

        external_parameters = external_parameters_ctx.get()
        if external_parameters is not None:
            self.name = external_parameters[0]
            self._parameters = external_parameters[1]

    @classmethod
    def _validate_passed_parameters_options(
        cls,
        name: str,
        parameters: Optional[ParametersType],
        parameters_class: Optional[type[ParametersType]],
    ) -> ParametersType:
        """
        Validates passed parameters and parameters class. If parameters
        passed then the instance will be used. If parameters class is passed,
        an attempt will be made to instantiate it. If neither parameters nor
        parameter class are passed, then the default base parameters will be
        used.

        Args:
            name: The name of the node.
            parameters: Parameters for the node.
            parameters_class: Parameters class.

        Returns:
            Validated parameters.

        Raises:
            ValueError: If parameters class instantiation fails.
        """
        if parameters is not None:
            if parameters_class is not None:
                logger.warning(
                    "Passed both parameters and parameters_class to the node "
                    f"'{name}'. Please use only parameters argument"
                )
            return parameters
        if parameters_class is None:
            fields = {
                name: copy(field)
                for name, field in NodeParameters.model_fields.items()
            }
            # Create subclass of NodeParameters. It's needed because otherwise
            # there will be an issue with type checking of subclasses.
            # For example: NodeRunSummary.parameters
            new_model = create_model(  # type: ignore
                NodeParameters.__name__,
                __doc__=NodeParameters.__doc__,
                __base__=NodeParameters,
                __module__=NodeParameters.__module__,
                **{
                    name: (info.annotation, info)
                    for name, info in fields.items()
                },
            )
            return cast(ParametersType, new_model())
        logger.warning(
            "parameters_class argument is deprecated. Please use "
            f"parameters argument for initializing node '{name}'."
        )
        try:
            return parameters_class()
        except ValidationError as e:
            raise ValueError(
                f"Can't instantiate parameters class of node '{name}'"
            ) from e

    def __copy__(self) -> "QualibrationNode[ParametersType]":
        """
        Creates a shallow copy of the node.

        This method copies the node, including its parameters, name, modes,
        and filepath, while resetting inspection-related modes to maintain
        consistency.

        Returns:
            A copy of the node.
        """
        modes = self.modes.model_copy(update={"inspection": False})
        instance = self.__class__(
            self.name, self.parameters_class(), self.description, modes=modes
        )
        instance.modes.inspection = self.modes.inspection
        instance.filepath = self.filepath
        return instance

    def copy(
        self, name: Optional[str] = None, **node_parameters: Any
    ) -> "QualibrationNode[ParametersType]":
        """
        Creates a modified copy of the node with updated parameters.

        The method allows the user to update parameters and assign a new name.
        If a new name is not provided, the copied node retains the same name as
        the original.

        Args:
            name: A new name for the node. Defaults to None.
            node_parameters: Additional parameters for the copied node.

        Returns:
            A copied node with the new parameters and name.

        Raises:
            ValueError: If the name provided is not a string.
        """
        logger.info(
            f"Copying node with name {self.name} with parameters "
            f"{name = }, {node_parameters = }"
        )
        if name is not None and not isinstance(name, str):
            raise ValueError(
                f"{self.__class__.__name__} should have a string name"
            )
        instance = self.__copy__()
        if name is not None:
            instance.name = name
        instance._parameters = instance.parameters_class.model_validate(
            node_parameters
        )
        instance.parameters_class = self.build_parameters_class_from_instance(
            instance._parameters
        )
        return instance

    def _warn_if_external_and_interactive_mpl(self) -> None:
        """
        Checks backend configuration and issues a warning if incompatible.

        Specifically, if the node is set to run in external mode and uses
        an interactive matplotlib backend, it switches to a non-interactive
        backend and logs a warning message.

        """
        mpl_backend = matplotlib.get_backend()
        if self.modes.external and mpl_backend in cast(
            list[str],
            backend_registry.list_builtin(BackendFilter.INTERACTIVE),  # type: ignore[no-untyped-call]
        ):
            matplotlib.use("agg")
            logger.warning(
                f"Using interactive matplotlib backend '{mpl_backend}' in "
                "external mode. The backend is changed to 'agg'."
            )

    def __str__(self) -> str:
        return f"{self.__class__.__name__}: {self.name}"

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}: {self.name} "
            f"(mode: {self.modes}; parameters: {self.parameters})"
        )

    @property
    def snapshot_idx(self) -> Optional[int]:
        """
        Returns the snapshot index from the storage manager.

        Retrieves the snapshot index that reflects the current state in
        the associated storage, if any.

        Returns:
            Snapshot index or None.
        """
        if self.storage_manager is None:
            return None
        return self.storage_manager.snapshot_idx

    def save(self) -> None:
        """
        Saves the current state of the node to the storage manager.

        If no storage manager is assigned, the method attempts to create
        a default configuration for storage and logs a warning message.

        Raises:
            ImportError: Raised if required configurations are not accessible.
        """
        if self.storage_manager is None:
            msg = (
                "Node.storage_manager should be defined to save node, "
                "resorting to default configuration"
            )
            logger.warning(msg)
            q_config_path = get_qualibrate_config_path()
            qs = get_qualibrate_config(q_config_path)
            state_path = get_quam_state_path(qs)
            self.storage_manager = LocalStorageManager(
                root_data_folder=qs.storage.location,
                active_machine_path=state_path,
            )
        self.storage_manager.save(
            node=cast("QualibrationNode[NodeParameters]", self)
        )

    def _load_from_id(
        self,
        node_id: int,
        base_path: Optional[Path] = None,
        custom_loaders: Optional[Sequence[type[BaseLoader]]] = None,
        build_params_class: bool = False,
    ) -> Optional["QualibrationNode[ParametersType]"]:
        """
        Loads a node by its identifier, parsing its content and data.

        Args:
            node_id: The unique identifier of the node.
            base_path: The base directory where node data is stored. If None,
                attempts to retrieve the path from qualibrate app settings.
            custom_loaders: An optional sequence of custom loader classes
                for handling specific file types.
            build_params_class: Whether to dynamically build a parameters class
                based on the node schema.

        Returns:
            The current `QualibrationNode` instance with the loaded data,
            or None if loading fails.
        """
        if base_path is None:
            q_config_path = get_qualibrate_config_path()
            qs = get_qualibrate_config(q_config_path)
            base_path = qs.storage.location
        node_dir = get_node_dir_path(node_id, base_path)
        if node_dir is None:
            logger.error(
                f"Node directory with id {node_id} wasn't found in {base_path}"
            )
            return None
        node_content = read_node_content(node_dir, node_id, base_path)
        if node_content is not None:
            quam_machine, parameters = parse_node_content(
                node_content,
                node_id,
                node_dir,
                build_params_class,
            )
            if quam_machine is not None:
                self.machine = quam_machine
            if parameters is not None:
                if build_params_class:
                    self.parameters_class = cast(
                        ParametersType, parameters
                    ).__class__
                    self._parameters = cast(ParametersType, parameters)
                else:
                    self._parameters = self.parameters.model_construct(
                        **cast(Mapping[str, Any], parameters)
                    )

        data = read_node_data(node_dir, node_id, base_path, custom_loaders)
        if data is not None:
            self.results = data
        return self

    @InstanceOrClassMethod
    def load_from_id(
        caller: Union[
            "QualibrationNode[ParametersType]",
            type["QualibrationNode[ParametersType]"],
        ],
        node_id: int,
        base_path: Optional[Path] = None,
        custom_loaders: Optional[Sequence[type[BaseLoader]]] = None,
    ) -> Optional["QualibrationNode[ParametersType]"]:
        """
        Class or instance method to load a node by its identifier.

        Args:
            caller: The class or instance calling this method. If called on
                a class, creates a new instance; otherwise, modifies the
                existing instance.
            node_id: The unique identifier of the node.
            base_path: The base directory where node data is stored. If None,
                attempts to retrieve the path from qualibrate app settings.
            custom_loaders: An optional sequence of custom loader classes
                for handling specific file types.

        Returns:
            A `QualibrationNode` instance with the loaded data, or None if
            loading fails.
        """
        instance: QualibrationNode[ParametersType] = (
            caller(name=f"loaded_from_id_{node_id}")
            if isinstance(caller, type)
            else caller
        )
        return instance._load_from_id(
            node_id=node_id,
            base_path=base_path,
            custom_loaders=custom_loaders,
            build_params_class=isinstance(caller, type),
        )

    def _post_run(
        self,
        last_executed_node: "QualibrationNode[ParametersType]",
        created_at: datetime,
        initial_targets: Sequence[TargetType],
        parameters: NodeParameters,
        run_error: Optional[RunError],
    ) -> NodeRunSummary:
        """
        Finalizes the node's execution and creates a summary.

        This method updates the outcomes for the last executed node,
        generates a summary, and logs the summary details. It is used
        to encapsulate the results of the node run, including the error
        (if any), state changes, and outcomes for each target.

        Args:
            last_executed_node: The node that was last executed.
            created_at: The timestamp when the run started.
            initial_targets: Targets at the start of the run.
            parameters: Parameters used in the run.
            run_error: Details of any error that occurred.

        Returns:
            A summary object containing execution details.
        """
        outcomes = last_executed_node.outcomes
        if self.parameters is not None and (targets := self.parameters.targets):
            lost_targets_outcomes = set(targets) - set(outcomes.keys())
            outcomes.update(
                {target: Outcome.SUCCESSFUL for target in lost_targets_outcomes}
            )
        self.outcomes = last_executed_node.outcomes = {
            name: Outcome(outcome) for name, outcome in outcomes.items()
        }
        self.run_summary = NodeRunSummary(
            name=self.name,
            description=self.description,
            created_at=created_at,
            completed_at=datetime.now(),
            initial_targets=initial_targets,
            parameters=parameters,
            outcomes=self.outcomes,
            error=run_error,
            successful_targets=[
                name
                for name, status in self.outcomes.items()
                if status == Outcome.SUCCESSFUL
            ],
            failed_targets=[
                name
                for name, status in self.outcomes.items()
                if status == Outcome.FAILED
            ],
            state_updates=self.state_updates,
        )
        logger.debug(f"Node run summary {self.run_summary}")
        return self.run_summary

    def run(
        self, interactive: bool = True, **passed_parameters: Any
    ) -> tuple["QualibrationNode[ParametersType]", BaseRunSummary]:
        """
        Runs the node with given parameters, potentially interactively.

        This function executes the node using parameters passed either
        directly or pre-configured in the node. It captures initial
        conditions, manages execution state, and logs errors or completion
        status.

        Args:
            interactive: Whether the node should be run interactively.
            **passed_parameters: Additional parameters to pass when
                running the node.

        Returns:
            The executed node and a summary of the run including outcomes,
            errors, and execution details.

        Raises:
            RuntimeError: Raised if the node filepath is not provided, or
                execution
        """
        logger.info(
            f"Run node {self.name} with parameters: {passed_parameters}"
        )
        if self.filepath is None:
            ex = RuntimeError(f"Node {self.name} file path was not provided")
            logger.exception("", exc_info=ex)
            raise ex
        params_dict = (
            self.parameters.model_dump() if self.parameters is not None else {}
        )
        params_dict.update(passed_parameters)
        parameters = self.parameters.model_validate(params_dict)
        initial_targets = copy(parameters.targets) if parameters.targets else []
        created_at = datetime.now()
        run_error: Optional[RunError] = None

        if run_modes_ctx.get() is not None:
            logger.error(
                "Run modes context is already set to %s",
                run_modes_ctx.get(),
            )
        run_modes_token = run_modes_ctx.set(
            RunModes(external=True, interactive=interactive, inspection=False)
        )
        external_parameters_token = external_parameters_ctx.set(
            (self.name, parameters)
        )
        try:
            self._parameters = parameters
            self.run_node_file(self.filepath)
        except Exception as ex:
            run_error = RunError(
                error_class=ex.__class__.__name__,
                message=str(ex),
                traceback=traceback.format_tb(ex.__traceback__),
            )
            logger.exception("", exc_info=ex)
            raise
        finally:
            run_modes_ctx.reset(run_modes_token)
            external_parameters_ctx.reset(external_parameters_token)
            last_executed_node = last_executed_node_ctx.get()
            if last_executed_node is None:
                logger.warning(
                    f"Last executed node not set after running {self}"
                )
                last_executed_node = self

            run_summary = self._post_run(
                last_executed_node,
                created_at,
                initial_targets,
                parameters,
                run_error,
            )

        return last_executed_node, run_summary

    def run_node_file(self, node_filepath: Path) -> None:
        """
        Executes the provided node file.

        This method runs the code in the given node file, ensuring that
        any interactive backends are temporarily disabled during the
        execution to avoid conflicts. Once the file has been executed,
        the original matplotlib backend is restored.

        Args:
            node_filepath: Path to the file that contains the node's execution
                logic.
        """
        mpl_backend = matplotlib.get_backend()
        # Appending dir with nodes can cause issues with relative imports
        try:
            matplotlib.use("agg")
            _module = import_from_path(
                get_module_name(node_filepath), node_filepath
            )
        finally:
            matplotlib.use(mpl_backend)

    def stop(self, **kwargs: Any) -> bool:
        """
        Halts the execution of the node if currently running.

        The method attempts to connect to the node's machine and stops
        the running job. If the necessary environment or conditions are
        not present, it will return False.

        Args:
            **kwargs: Additional keyword arguments that might be used for
                stopping the node.

        Returns:
            True if the node is successfully stopped, False otherwise.
        """
        active_node = self.__class__.active_node
        logger.debug(f"Stop node {active_node} with id {id(active_node)}")
        if active_node is None:
            return False
        logger.info(f"Stop node {active_node.name}")
        if find_spec("qm") is None:
            return False
        qmm = getattr(active_node.machine, "qmm", None)
        if not qmm:
            if active_node.machine is None or not hasattr(
                active_node.machine, "connect"
            ):
                return False
            qmm = active_node.machine.connect()
        if hasattr(qmm, "list_open_qms"):
            ids = qmm.list_open_qms()
        elif hasattr(qmm, "list_open_quantum_machines"):
            ids = qmm.list_open_quantum_machines()
        else:
            return False
        qm = qmm.get_qm(ids[0])
        job = qm.get_running_job()
        if job is None:
            return False
        job.halt()
        return True

    @contextmanager
    def record_state_updates(
        self, interactive_only: bool = True
    ) -> Generator[None, None, None]:
        """
        Records state updates for the node during execution.

        This method wraps around node operations to record all state updates
        that occur, specifically during interactive execution. It uses
        custom `__setattr__` and `__setitem__` functions for relevant classes
        to record these changes.

        Args:
            interactive_only: Whether to only record in interactive mode.
                Defaults to True.

        Yields:
            None: Allows wrapped operations to execute while recording state
                updates.
        """
        if not self.modes.interactive and interactive_only:
            yield
            return

        logger.debug(f"Init recording state updates for node {self.name}")
        # Override QuamComponent.__setattr__()
        try:
            from quam.core import (
                QuamBase,
                QuamComponent,
                QuamDict,
                QuamList,
                QuamRoot,
            )
        except ImportError as ex:
            print(ex)
            yield
            return

        quam_classes_mapping = (
            QuamBase,
            QuamComponent,
            QuamRoot,
            QuamDict,
        )
        quam_classes_sequences = (QuamList, QuamDict)

        cls_setattr_funcs = {
            cls: cls.__dict__["__setattr__"]
            for cls in quam_classes_mapping
            if "__setattr__" in cls.__dict__
        }
        cls_setitem_funcs = {
            cls: cls.__dict__["__setitem__"]
            for cls in quam_classes_sequences
            if "__setitem__" in cls.__dict__
        }
        try:
            for cls in cls_setattr_funcs:
                cls.__setattr__ = partialmethod(
                    record_state_update_getattr, node=self
                )
            for cls in cls_setitem_funcs:
                cls.__setitem__ = partialmethod(
                    record_state_update_getitem, node=self
                )
            yield
        finally:
            for cls, setattr_func in cls_setattr_funcs.items():
                cls.__setattr__ = setattr_func
            for cls, setitem_func in cls_setitem_funcs.items():
                cls.__setitem__ = setitem_func

    @classmethod
    def scan_folder_for_instances(
        cls, path: Path
    ) -> dict[str, QRunnable[ParametersType, NodeRunParametersType]]:
        """
        Scans a directory for node instances and returns them.

        This method scans a folder to locate all node files that are valid
        instances of `QualibrationNode`. It sets an inspection mode to avoid
        executing the nodes during scanning.

        Args:
            path: The directory to scan for node files.

        Returns:
            A dictionary of node names to their corresponding node instances.
        """
        nodes: dict[str, QRunnable[ParametersType, NodeRunParametersType]] = {}
        if run_modes_ctx.get() is not None:
            logger.error(
                "Run modes context is already set to %s",
                run_modes_ctx.get(),
            )
        run_modes_token = run_modes_ctx.set(RunModes(inspection=True))
        try:
            for file in sorted(path.iterdir()):
                if not file_is_calibration_instance(file, cls.__name__):
                    continue
                try:
                    cls.scan_node_file(file, nodes)
                except Exception as e:
                    logger.warn(
                        "An error occurred on scanning node file "
                        f"{file.name}.\nError: {type(e)}: {e}"
                    )

        finally:
            run_modes_ctx.reset(run_modes_token)
        return nodes

    @classmethod
    def scan_node_file(
        cls,
        file: Path,
        nodes: dict[str, QRunnable[ParametersType, NodeRunParametersType]],
    ) -> None:
        """
        Scans a node file and adds its instance to the provided dictionary.

        This method scans the content of a given file to identify if it
        contains a valid `QualibrationNode`. If so, it adds the node to
        the given nodes dictionary for further processing.

        Args:
            file: The node file to scan.
            nodes: dictionary to add valid nodes to.

        Raises:
            StopInspection: Used to stop execution once inspection completes.
        """
        logger.info(f"Scanning node file {file}")
        try:
            # TODO Think of a safer way to execute the code
            _module = import_from_path(get_module_name(file), file)
        except StopInspection as ex:
            node = cast("QualibrationNode[ParametersType]", ex.instance)
            node.filepath = file
            node.modes.inspection = False
            cls.add_node(node, nodes)

    @classmethod
    def add_node(
        cls,
        node: "QualibrationNode[ParametersType]",
        nodes: dict[str, QRunnable[ParametersType, NodeRunParametersType]],
    ) -> None:
        """
        Adds a node instance to the node dictionary.

        If a node with the same name already exists in the dictionary,
        this method overwrites the existing entry with a warning.

        Args:
            node: The node instance to add.
            nodes: dictionary to store nodes.
        """
        if node.name in nodes:
            logger.warning(
                f'Node "{node.name}" already exists in library, overwriting'
            )

        nodes[node.name] = node


if __name__ == "__main__":
    from pathlib import Path

    # path = Path("/home/maxim_v4s/Downloads/")
    path = Path.home().joinpath(".qualibrate/user_storage/init_project")
    node = QualibrationNode.load_from_id(12)
    print(node.results)
    print(node.parameters)



from importlib.util import module_from_spec, spec_from_file_location
from pathlib import Path
from types import ModuleType


def get_module_name(file_path: Path) -> str:
    """Create module name from file path."""
    return f"_node_{file_path.stem}"


def import_from_path(module_name: str, file_path: Path) -> ModuleType:
    """Import a module given its name and file path."""
    spec = spec_from_file_location(module_name, file_path)
    if spec is None:
        raise RuntimeError(f"Can't read spec from {file_path}")
    module = module_from_spec(spec)
    if spec.loader is None:
        raise RuntimeError(f"Can't get loader from spec of file {file_path}")
    spec.loader.exec_module(module)
    return module



from typing import Any

from qualibrate.q_runnnable import (
    CreateParametersType,
    QRunnable,
    RunParametersType,
)


class StopInspection(Exception):
    def __init__(
        self,
        *args: Any,
        instance: QRunnable[CreateParametersType, RunParametersType],
    ):
        super().__init__(*args, instance)
        self.instance = instance



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Generic, Optional, TypeVar

if TYPE_CHECKING:
    from qualibrate.parameters import NodeParameters
    from qualibrate.qualibration_node import QualibrationNode


NodeTypeVar = TypeVar("NodeTypeVar", bound="QualibrationNode[NodeParameters]")


class StorageManager(ABC, Generic[NodeTypeVar]):
    """
    Abstract base class for managing the storage of calibration nodes.

    The `StorageManager` class provides an interface for saving the state of
    calibration nodes (`QualibrationNode`). It includes functionality for
    managing snapshots of the node's state.
    """

    snapshot_idx: Optional[int] = None

    @abstractmethod
    def save(self, node: NodeTypeVar) -> None:
        """
        Saves the current state of the provided node.

        This method is an abstract method, and should be implemented by any
        subclass to provide the logic for saving the state of a
        `QualibrationNode`.

        Args:
            node (QualibrationNode): The node whose state is to be saved.
        """
        pass



import logging
from pathlib import Path

if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    from qualibrate import NodeParameters, QualibrationLibrary, QualibrationNode

    library_folder = Path(__file__).parent.parent / "calibrations"
    library = QualibrationLibrary[QualibrationNode[NodeParameters]](
        library_folder=library_folder
    )

    print(library.nodes)
